import numpy
from .splitter import *

class Attribute():
    def __init__(self, index=None, name=''):
        if index is None: raise ValueError('must pass an index')
        self.index = index
        self.name = name

    def each_splitter(self, values):
        raise NotImplementedError

class CategoricalAttribute(Attribute):
    def each_splitter(self, values):
        unique_values = numpy.unique(values)
        if len(unique_values) > 1:
            for value in unique_values:
                yield IsEqualSplitter(self, value)

class NumericAttribute(Attribute):
    def each_splitter(self, values):
        stddev, maximum, value = values.std(), values.max(), values.min()
        while maximum > value:
            yield GreaterThanOrEqualToSplitter(self, value)
            value += stddev / 10.0 # TODO: this is arbitrary?



from collections import Counter
from math import log

class OutcomeCounter():
    def __init__(self, outcomes=[]):
        self.counter = Counter()
        self.total = 0
        for outcome in outcomes:
            self.record(outcome)

    def record(self, outcome):
        self.counter[outcome] += 1
        self.total += 1

    def entropy_of(self, outcome):
        probability = self[outcome] / float(self.total)
        return -probability * log(probability, 2)

    def outcomes(self):
        return self.counter.keys()

    def entropy(self):
        return sum(map(self.entropy_of, self.outcomes()))

    def most_common_value(self):
        return self.counter.most_common(1)[0][0]

    def __len__(self):
        return len(self.counter)

    def __getitem__(self, outcome):
        return self.counter[outcome]



from collections import defaultdict
from collections import Counter
from .outcome_counter import *
from .attribute import *
from .single_attribute_splitter_finder import *
import random
import numpy

class Dataset():
    def __init__(self, X, y, attributes=None):
        self.X = X
        self.y = y
        self.outcome_counter = OutcomeCounter(y)
        self.attributes = attributes or [CategoricalAttribute(i) for i in range(X.shape[1])]
        assert self.X.shape[0] == len(y), "len(y) must match len(X)"
        assert self.X.shape[1] == len(self.attributes), "len(attributes) must match len(X[i])"

    def __len__(self):
        return self.X.shape[0]

    def best_single_attribute_splitter(self):
        finder = SingleAttributeSplitterFinder(self, n=len(self.attributes))
        return finder.best_splitter()

    def splitter_entropy(self, splitter):
        splits = defaultdict(OutcomeCounter)
        for i in range(len(self)):
            splits[splitter.split(self.X[i])].record(self.y[i])
        return sum(y.total * y.entropy() for y in splits.values()) / float(len(self))

    def split_on(self, splitter):
        splits = defaultdict(list)
        for i in range(len(self)):
            splits[splitter.split(self.X[i])].append(i)
        return { value: self.take(indices) for value, indices in splits.items() }

    def take(self, indices):
        return self.__class__(self.X.take(indices, 0), self.y.take(indices), self.attributes)

    def random_split(self, fraction):
        n_examples = int(len(self) * fraction)
        indices = list(range(len(self)))
        random.shuffle(indices)
        return self.take(indices[:n_examples]), self.take(indices[n_examples:])

    def bootstrap(self, n=None):
        return self.take([random.randrange(len(self)) for _i in range(n or len(self))])



from collections import Counter
from .classifier import *
from .decision_tree import *

class RandomForest(Classifier):
    def __init__(self, dataset, tree_class=DecisionTree, n_trees=10, examples_per_tree=None):
        self.trees = [tree_class(dataset.bootstrap(examples_per_tree)) for _i in range(n_trees)]

    def vote_on(self, x):
        # TODO: we could return early as soon as we have a definite plurality
        return Counter(tree.classify(x) for tree in self.trees)

    def classify(self, x):
        return self.vote_on(x).most_common(1)[0][0]



class Splitter():
    def split(self, x):
        raise NotImplementedError

class SingleAttributeSplitter(Splitter):
    def __init__(self, attribute, value):
        self.attribute = attribute
        self.value = value

class IsEqualSplitter(SingleAttributeSplitter):
    def split(self, x):
        return x[self.attribute.index] == self.value

class GreaterThanOrEqualToSplitter(SingleAttributeSplitter):
    def split(self, x):
        return x[self.attribute.index] >= self.value



import random
import numpy

class SingleAttributeSplitterFinder():
    def __init__(self, dataset, n=None):
        self.dataset = dataset
        if n is None:
            n = int(numpy.ceil(numpy.sqrt(len(dataset.attributes))))
        self.attributes = random.sample(dataset.attributes, n)

    def each_splitter(self):
        for attribute in self.attributes:
            attr_values = self.dataset.X[:, attribute.index]
            for splitter in attribute.each_splitter(attr_values):
                yield splitter

    def best_splitter(self):
        return min(self.each_splitter(), key=self.dataset.splitter_entropy,
                   default=None)



from .classifier import *
from .single_attribute_splitter_finder import *

class DecisionTree(Classifier):
    def __init__(self, dataset,
                 min_samples_split=2,
                 max_depth=float('inf'),
                 depth=1,
                 splitter_finder=SingleAttributeSplitterFinder):
        self.depth = depth
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.outcome_counter = dataset.outcome_counter
        self.splitter = None
        self.branches = {}

        # Stop recursing if we're at a leaf
        if depth >= max_depth: return
        if len(dataset) < min_samples_split: return
        if len(self.outcome_counter) == 1: return

        # Otherwise, branch
        splitter = splitter_finder(dataset).best_splitter()
        if splitter:
            self.splitter = splitter
            self.branches = { value: self.new_branch(subset)
                for value, subset in dataset.split_on(splitter).items() }

    def new_branch(self, dataset):
        return self.__class__(dataset,
                min_samples_split=self.min_samples_split,
                max_depth=self.max_depth,
                depth=self.depth+1)

    def splitter_finder(self, dataset):
        raise NotImplementedError

    def branch_for(self, x):
        # we are a leaf node unless we have a splitter
        if self.splitter:
            value = self.splitter.split(x)
            # if we have a missing value (impossible for binary splits, but
            # possible for other subclasses), return most common value
            if value in self.branches:
                return self.branches[value]
        return None

    def leaf_value(self):
        return self.outcome_counter.most_common_value()

    def classify(self, x):
        branch = self.branch_for(x)
        if branch:
            return branch.classify(x)
        else:
            return self.leaf_value()



# -*- coding: utf-8 -*-

from collections import defaultdict
import numpy
from .outcome_counter import *

class ConfusionMatrix:
    #            Real
    #        y1  y2  y3  y4
    #Pred y1  5   1   0   2
    #
    #     y2  2  10   3   1
    #
    #     y3  8   1   5   0
    #
    #     y4  0   0   0   5
    def __init__(self, predictions, realities):
        self.counts = defaultdict(OutcomeCounter)
        outcomes = set()
        n_correct = 0
        self.total = 0
        for i in range(len(realities)):
            yr = realities[i]
            yp = predictions[i]
            outcomes |= set([yr, yp])
            n_correct += yr == yp
            self.counts[yr].record(yp)
            self.total += 1
        self.accuracy = n_correct / float(self.total)
        self.outcomes = sorted(list(outcomes))

    def to_array(self):
        return numpy.array([[self.counts[o][oo] for oo in self.outcomes] for o in self.outcomes])

    def true_positives(self, outcome):
        return self.counts[outcome][outcome]

    def true_negatives(self, outcome):
        return sum(self.counts[o].total - self.counts[o][outcome]
            for o in self.outcomes if o != outcome)

    def false_positives(self, outcome):
        return sum(self.counts[o][outcome]
            for o in self.outcomes if o != outcome)

    def false_negatives(self, outcome):
        return self.counts[outcome].total - self.counts[outcome][outcome]

    def sensitivity_for(self, outcome):
        # probability of predicting outcome given real value is outcome
        # high sensitivity =>
        #   classifier rules out outcome if we didn't predict it
        return self.true_positives(outcome) / float(self.counts[outcome].total)

    def specificity_for(self, outcome):
        # probability of predicting ¬outcome given real value is ¬outcome
        # high specificity =>
        #   classifier rules in outcome if we did predict it
        return self.true_negatives(outcome) / float(self.total - self.counts[outcome].total)

    def sensitivity(self):
        return numpy.mean([self.sensitivity_for(o) for o in self.outcomes])

    def specificity(self):
        return numpy.mean([self.specificity_for(o) for o in self.outcomes])



from .confusion_matrix import *

class Classifier():
    def __init__(self, training_dataset):
        raise NotImplementedError

    def classify(self, point):
        raise NotImplementedError

    def performance_on(self, test_dataset):
        predictions = [self.classify(x) for x in test_dataset.X]
        realities = test_dataset.y
        return ConfusionMatrix(predictions, realities)



import setuptools
import oo_trees

setuptools.setup(
    name='oo_trees',
    version=oo_trees.__version__,
    test_suite='tests',
    url='http://github.com/asross/oo_trees',
    license='Apache Software License',
    author='Andrew Ross',
    install_requires=[
      'numpy>=1.10.4',
      'scikit-learn>=0.17',
      'scipy>=0.16.1'
    ],
    author_email='andrewslavinross@gmail.com',
    description='Object-oriented implementations of decision tree variants',
    packages=['oo_trees'],
    include_package_data=True,
    platforms='any',
    classifiers=[
      'Programming Language :: Python',
      'Development Status :: 3 - Alpha',
      'Natural Language :: English',
      'License :: OSI Approved :: Apache Software License',
      'Operating System :: OS Independent',
      'Intended Audience :: Developers',
      'Intended Audience :: Science/Research'
    ]
)
